/*
  Copyright 2006,2008 CNRS

  Authors: Mathieu Poirier (CNRS), Florent Lamiraux (CNRS)

*/


/*****************************************
 INCLUDES
*******************************************/
#include <sys/time.h>
#include <libgen.h>

#include <dynamicsJRLJapan/dynamicsJRLJapanFactory.h>
#include "KineoWorks2/kwsJointFreeFlyer.h"
#include "hppWalkPatGen.h"
#include "hppModel/hppDevice.h"
#include "hrp2_14/hrp2_14.h"
#include "kwsIO/kwsioConfig.h"

/*****************************************
 DEFINES
*******************************************/
#define HRP2_TIME_STEP 0.005

// Select level of verbosity at configuration.
#if DEBUG==3
#define ODEBUG3(x) std::cout << "ChppWalkPatGen:" << x << std::endl
#define ODEBUG2(x) std::cout << "ChppWalkPatGen:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkPatGen:" << x << std::endl
#elif DEBUG==2
#define ODEBUG3(x)
#define ODEBUG2(x) std::cout << "ChppWalkPatGen:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkPatGen:" << x << std::endl
#elif DEBUG==1
#define ODEBUG3(x)
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "ChppWalkPatGen:" << x << std::endl
#else
#define ODEBUG3(x)
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

/*****************************************
 METHODS
*******************************************/


// ==========================================================================

ChppWalkPatGen::ChppWalkPatGen(double inStepLength, double inStepSpace) :
  attStepLength(inStepLength), attStepSpacing(inStepSpace),
  attStringStream()
{
  double temp[40] =  HALFSITTINGPOSITION_DEG_OPENHRP ;
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = temp[i]*M_PI/180.0;
  }
}

// ==========================================================================

void ChppWalkPatGen::reset()
{
    //Set initial Position - half sitting position
  MAL_VECTOR_DIM(initialPosition,double,40);

  double temp[40] =  HALFSITTINGPOSITION_DEG_OPENHRP ;
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = temp[i]*M_PI/180.0;
    initialPosition(i) =  attInitOpenHRPPosturePatGen[i];
  }
  attWalkPatGen->SetCurrentJointValues(initialPosition);
}

// ==========================================================================

ChppWalkPatGen::~ChppWalkPatGen()
{
  delete attWalkPatGen;
}

// ==========================================================================


ktStatus ChppWalkPatGen::initOpenHRPPosturePatGenRad(double initPos[40])
{
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = initPos[i];
  }
  return KD_OK;
}

// ==========================================================================

void ChppWalkPatGen::initWalkPatGen(CjrlHumanoidDynamicRobot* inRobot)
{
  // Create Pattern Generator Interface
  attWalkPatGen =patternGeneratorInterfaceFactory(inRobot);

  const char lBuffer[9][256] =
    {":omega 0.0",
     ":stepheight 0.07",
     ":singlesupporttime 0.78",
     ":doublesupporttime 0.02",
     ":armparameters 0.5",
     ":LimitsFeasibility 0.0",
     ":ZMPShiftParameters 0.015 0.015 0.015 0.015",
     ":TimeDistributionParameters 2.0 3.7 1.7 3.0",
     ":UpperBodyMotionParameters -0.1 -1.0 0.0"
    };

  for(int i=0;i<9;i++) {
    std::istringstream strm(lBuffer[i]);
    attWalkPatGen->ParseCmd(strm);
  }
}

// ==========================================================================

void ChppWalkPatGen::computeStreamFromStepSequence(const ChppWalkStepSequence& inRelativeFootPrintVector)
{
  std::istringstream strm2 ;
  std::stringstream strm ;

  strm << ":stepseq " ;
  for (unsigned int count =0 ;count < inRelativeFootPrintVector.size() ; count++) {
    strm <<"\t" << inRelativeFootPrintVector[count].sx
	 <<"\t" << inRelativeFootPrintVector[count].sy
	 <<"\t" << inRelativeFootPrintVector[count].theta;
  }

  strm2.str(strm.str()) ;

  attWalkPatGen->ParseCmd(strm2);
}

// ==========================================================================

ktStatus ChppWalkPatGen::computeAllQFromAStepSequence(ChppDeviceShPtr hppDevice, CkwsPathShPtr& outPathHumanoid,
						      std::vector<CkwsConfigShPtr>& outDiscretizedPath, std::vector<double> &HRP2TimeRef,
						      std::vector<std::vector<double> > &HRP2ZMPRef, std::vector<std::vector<double> > &HRP2WaistPosition,
						      const ChppWalkStepSequence& inRelativeFootPrintVector)
{
  // collision status
  ktStatus col_state = KD_OK;

  CkwsConfigShPtr startConfig=CkwsConfig::create(hppDevice) ;
  CkwsSMLinearShPtr steeringMethod = CkwsSMLinear::create();
  bool firstIteration = true ;

  double xInit, yInit, zInit, thetaInit;

  inRelativeFootPrintVector.getInitialPosition(xInit, yInit, zInit, thetaInit);

  // Set the walking mode:
  //     - 0 : Normal mode.
  //     - 1 : Waist height variation.
  //     - 3 : Kineo works mode.
  //     - 4 : Following the upper-body motion provided by another plugin.
  // Specify the walking mode: here the default one.
  std::istringstream strm2(":walkmode 0");
  attWalkPatGen->ParseCmd(strm2);


  computeStreamFromStepSequence(inRelativeFootPrintVector) ;

  unsigned long int NbOfIt=0;

  MAL_VECTOR_DIM(currentAcceleration,double,46);
  MAL_VECTOR_DIM(currentVelocity,double,46);
  MAL_VECTOR_DIM(currentConfiguration,double,46);
  MAL_VECTOR_DIM(ZMPTarget,double,3);



  // Main Loop of PatternGenerator Computation
  while(attWalkPatGen->RunOneStepOfTheControlLoop(currentConfiguration,
						  currentVelocity,
						  currentAcceleration,
						  ZMPTarget)) {

    NbOfIt++;

    //
    // WalkGenJrl defines motions in the reference frame of the
    // initial configuration of the robot
    // We need to express the waist position in world frame.
    //

    double xPosShifted, yPosShifted, zPosShifted, rollPosShifted, pitchPosShifted, yawPosShifted ;
    xPosShifted = currentConfiguration(0)*cos(thetaInit) - currentConfiguration(1)*sin(thetaInit) + xInit ;
    yPosShifted = currentConfiguration(0)*sin(thetaInit) + currentConfiguration(1)*cos(thetaInit) + yInit ;
    zPosShifted = currentConfiguration(2)+ zInit ;
    rollPosShifted = currentConfiguration(3) ;
    pitchPosShifted = currentConfiguration(4) ;
    yawPosShifted = currentConfiguration(5) + thetaInit;

    currentConfiguration(0) = xPosShifted ;
    currentConfiguration(1) = yPosShifted ;
    currentConfiguration(2) = zPosShifted ;
    currentConfiguration(3) = rollPosShifted ;
    currentConfiguration(4) = pitchPosShifted ;
    currentConfiguration(5) = yawPosShifted ;

    // ------------------------ Write in Data Structures ---------------//

    std::vector<double> currentZMPTarget(3) ;
    std::vector<double> currentWaistPosition(6) ;
    std::vector<double> currentOpenHrpConfig(40) ;

    double time = HRP2_TIME_STEP*(HRP2TimeRef.size()+1);
    HRP2TimeRef .push_back(time) ;
    ODEBUG2(" HRP2TimeRef size= " <<HRP2TimeRef.size() << "; time = " << time);
    currentZMPTarget[0] = ZMPTarget(0) ; currentZMPTarget[1] = ZMPTarget(1) ; currentZMPTarget[2] = ZMPTarget(2) ;
    HRP2ZMPRef.push_back(currentZMPTarget) ;
    ODEBUG2(" HRP2ZMPRef  size= " <<  HRP2ZMPRef.size()  );

    //
    // Record position of waist along motion
    //
    for (unsigned int i=0; i<5; i++) {
      currentWaistPosition[i] = currentConfiguration(i);
    }

    HRP2WaistPosition.push_back(currentWaistPosition) ;
    ODEBUG2(" HRP2WaistPosition size= " << HRP2WaistPosition.size()  );
    // Be careful do not insert currentConfiguration(0)->currentConfiguration(5) :
    // because they are position and orientation
    for( unsigned int count = 6 ; count <  MAL_VECTOR_SIZE(currentConfiguration) ; count++) {
      currentOpenHrpConfig[count-6] = currentConfiguration(count) ;
    }

    //
    // Express configuration in KineoWorks system of coordinates
    //
    CkwsConfigShPtr endConfig  = CkwsConfig::create(hppDevice);
    std::vector<double> robotDofValues(hppDevice->countDofs());

    if (!hppDevice->jrlDynamicsToKwsDofValues(currentConfiguration, robotDofValues)) {
      std::cout << *hppDevice << std::endl;
      throw std::string("Failed to convert config from jrlDynamics to KineoWorks.");
    }
    endConfig->setDofValues(robotDofValues);
    outDiscretizedPath.push_back(endConfig) ;
    ODEBUG2(" outDiscretizedPath size= " <<  outDiscretizedPath.size()  );



    // -----------------------------------------------------------------------------//

    if (firstIteration) {
      /*
	At the first iteration, the initial configuration is set to
	  - the end configuration if the current path is empty
          - the end configuration of the current path otherwise
      */
      if (outPathHumanoid->isEmpty()) {
	for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	  startConfig->dofValue(z, endConfig->dofValue(z)) ;
	}
      }
      else {
	startConfig = outPathHumanoid->configAtEnd();
      }

      firstIteration = false ;
    } else {

      // note : Kws can't built a Direct Path with startConfig = endConfig
      //creation of the direct path
      CkwsDPLinearShPtr DPLinear = CkwsDPLinear::create(*startConfig, *endConfig, steeringMethod) ;

      // If the robot has not moved, no path will be created, in that case, skip
      // the end of the loop.
      if (!DPLinear)
	continue;

      // check if the path is collising once for 2 configuration
      if(NbOfIt % 2 == 0){
	hppDevice->setCurrentConfig(*endConfig);
	if(hppDevice->collisionTest()){
	  cout<<"(self-)collision detected at t="<<NbOfIt*0.005<<endl;
#if DEBUG < 2
	  outPathHumanoid.reset ();
#endif
	  return KD_ERROR;
	}
      }

      if (!DPLinear) {
	ODEBUG1(":computeAllQFromAStepSequence: failed to create direct path");
	ODEBUG1(":computeAllQFromAStepSequence: startConfig=("
		<< *startConfig << ")");
	ODEBUG1(":computeAllQFromAStepSequence: endConfig=("
		<< *endConfig << ")");
#if DEBUG < 2
	outPathHumanoid.reset ();
#endif
	return KD_ERROR;
      }

      //add the direct path to the path
      if (outPathHumanoid->appendDirectPath(DPLinear) == KD_ERROR) {
	ODEBUG1(":computeAllQFromAStepSequence: failed to insert direct path.");
	ODEBUG1(":computeAllQFromAStepSequence: config at end of path = ("
		<< *(outPathHumanoid->configAtEnd()) << ")");
	ODEBUG1(":computeAllQFromAStepSequence: config at start of direct path = ("
		<< *(DPLinear->configAtStart()) << ")");
#if DEBUG < 2
	outPathHumanoid.reset ();
#endif
	return KD_ERROR;
      }
      ODEBUG2(" outPathHumanoid length = " << outPathHumanoid->countDirectPaths());
      for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	startConfig->dofValue(z, endConfig->dofValue(z)) ;
      }
    }
  }

  ODEBUG1("");
  ODEBUG1(" number of iterations " << NbOfIt);

  return col_state;
}

