#include <iostream>
#include <fstream>

#include "jrlMathTools/angle.h"

#include "testDirectPath.h"
#include "kwsIO/kwsioConfig.h"

#include "KineoWorks2/kwsSteeringMethod.h"
#include "KineoModel/kppSolidComponent.h"
#include "KineoModel/kppSolidComponentRef.h"
#include "KineoModel/kppFreeFlyerJointComponent.h"
#include "KineoKCDModel/kppKCDBox.h"

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "CtestDirectPath:" << x << std::endl
#define ODEBUG1(x) std::cerr << "CtestDirectPath:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "CtestDirectPath:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

/*--------------------------------------------------
 *
 *                P U B L I C    M E T H O D S
 *
 *-------------------------------------------------- */


ktStatus CtestDirectPath::init()
{
  if (createDevice() != KD_OK) {
    ODEBUG1(":init: Could not create device.");
    return KD_ERROR;
  }

  if (createSteeringMethod() != KD_OK) {
    ODEBUG1(":init: Could not create steering method.");
    return KD_ERROR;
  }

  return KD_OK;
}

/*--------------------------------------------------
 *
 *                P R I V A T E    M E T H O D S
 *
 *-------------------------------------------------- */

ktStatus CtestDirectPath::createDevice()
{
  double footWidth=0.2;
  // Create empty device and store it in object
  CkppDeviceComponentShPtr kppDevice=CkppDeviceComponent::create("HBB");

  CkitMat4 absPos;
  CkppKCDBoxShPtr polyBox = CkppKCDBox::create("HBB-Box",0.5420, 0.8682, 1.6909);
  polyBox->diffuseColor( CkppColor( 0.2f, 0.2f, 0.6f, 0.4f ) );
  absPos.translate(0.0423,0.0,0.7670);
  polyBox->setAbsolutePosition( absPos );
  polyBox->makeCollisionItem();
  
  
  // create ExtraDof
  CkwsDofShPtr kwsExtraDof ;
  kwsExtraDof = CkwsDof::create(false);
  // Bound curvature
  kwsExtraDof->bounds(-5.0, 5.0);
  kwsExtraDof->isBounded(true);
  kppDevice->addExtraDof(kwsExtraDof);
  
  
  // build plan root joint
  CkppJointComponentShPtr kppJoint = CkppFreeFlyerJointComponent::create( "ROOTJOINT.BB" );
  // Bound joint
  // X
  kppJoint->kwsJoint()->dof(0)->isBounded(true);
  kppJoint->kwsJoint()->dof(0)->bounds(-10.0,10.0); 
  // Y
  kppJoint->kwsJoint()->dof(1)->isBounded(true);
  kppJoint->kwsJoint()->dof(1)->bounds(-10.0,10.0); 
  // Z
  kppJoint->kwsJoint()->dof(2)->isLocked(true);
  // Rx
  kppJoint->kwsJoint()->dof(3)->isLocked(true);
  // Ry
  kppJoint->kwsJoint()->dof(4)->isLocked(true);
  // Rz
  kppJoint->kwsJoint()->dof(5)->isBounded(false);
  kppJoint->kwsJoint()->dof(5)->bounds(-M_PI,M_PI);

  kppJoint->kwsJoint()->dof(0)->userWeightScaler(2);
  kppJoint->kwsJoint()->dof(1)->userWeightScaler(2);
  kppJoint->kwsJoint()->dof(5)->userWeightScaler(2*footWidth/0.854365);

  //attach the body to the joint 
  kppJoint->addSolidComponentRef( CkppSolidComponentRef::create(polyBox) ); 
  kppJoint->doesDisplayPath( true );
  
  //attach the joint to the device
  kppDevice->rootJointComponent(kppJoint);

  // create to config for test
  CkwsConfigShPtr startConfig=CkwsConfig::create(kppDevice->kwsDevice()) ;
  CkwsConfigShPtr endConfig  =CkwsConfig::create(kppDevice->kwsDevice()) ;

  startConfig->dofValue(0, 0.0) ;  // Kappa
  startConfig->dofValue(1, 0.0) ;  // x  
  startConfig->dofValue(2, 0.0) ;  // y
  startConfig->dofValue(6, 0.0) ;  // theta
  
  endConfig->dofValue(0, 0.0) ; // Kappa  
  endConfig->dofValue(1, 2.0) ; // x
  endConfig->dofValue(2, 1.0) ; // y
  endConfig->dofValue(6, 0.0) ; // theta 
  
  attDevice = kppDevice;
  
  return KD_OK;
}

ktStatus CtestDirectPath::createSteeringMethod()
{
  // Create steering method : Flat Interpolation 
  CkwsSteeringMethodShPtr kwsSteeringMethod = CkwsSMLinear::create(false);
  if (!kwsSteeringMethod) {
    return KD_ERROR;
  }
  attSteeringMethod = kwsSteeringMethod;
  return KD_OK;
}

